#ifndef __colorbar_h__
#define __colorbar_h__

#include "widget.h"
#include "colormap.h"
#include "itt.h"

// Colormap(s) are X window Colormap ID's. ColorMapInfo(s) contain a fixed number of
// real RGB color cells. RGB Values of ColorMapInfo(s) are never altered.
// Colormap(s) are initialized from ColorMapInfo values. Colormap RGB values will
// changed, based on user input.
// ITTs (Intensity Transfer Table) are used by RTD based colormaps only.

// Colorbar

class Colorbar : public Widget {
protected:
  unsigned short* colorIndex; // array of allocated color index
  unsigned char* colorCells; // array of allocated color values
  ColorMapInfo* cmaps;       // list of known colormaps
  ColorMapInfo* currentcmap; // current colormap;
  ITT* itts;                 // list of known itts
  ITT* currentitt;           // current itt;

  float bias;                // current bias (default = 0.0)
  float contrast;            // current contrast (default = 1.0)
  int invert;                // flag for inverted colormap

  int colorCount;            // number of dynamic color cells allocated

protected:
  virtual int initColormap() =0;        // initialize current colormap
  virtual void updateColors() =0;       // ColorMapInfo to Colormap
  virtual int updatePixmap(const BBox&) =0; // renders image/graphics into pixmap

  void loadDefaultCMaps();   // load default ColorMaps
  int calcContrastBias(int); // map index based on current contrast/bias

private:
  void reset();              // reset current colormap
  ColorMapInfo* newColorMap(const char*, const char*); // create new cmap
  ITT* newITT(const char*);  // create new cmap based on file name

public:
  Colorbar(Tcl_Interp*, Tk_Canvas, Tk_Item*);
  virtual ~Colorbar();

  int parse(istrstream&);            // parse subcommands

  // SubCommandFunctions

  int configure(int, char* [], int); // parse config options
  void adjustCmd(float, float);    // set current constast/bias
  void getColormapCmd();           // get current colormap, bias, contrast, colors
  void getColormapNameCmd(int);    // get colormap name for id
  void getCurrentNameCmd();        // return name of current colormap
  void getCurrentIDCmd();          // return id of current colormap
  void getCurrentFileNameCmd();    // return file name of current colormap
  void getInvertCmd();             // return current status of invert
  void invertCmd(int);             // invert current colormap
  void ittCmd(const char*);        // set current itt
  void ittCmd(int);                // set current itt
  void listIDCmd();                // list loaded colormaps
  void listNameCmd();              // list loaded colormaps
  void listITTIDCmd();             // list loaded itts
  void listITTNameCmd();           // list loaded itts
  void loadCmd(const char*, const char*); // load new colormap
  void loadITTCmd(const char*);    // load new itt
  void mapCmd(char*);              // set current colormap
  void mapCmd(int);                // set current colormap
  void saveCmd(const char*);       // save current colormap
  void setColormapCmd(char*, float, float, int); // set current colormap
  void setColormapCmd(int, float, float, int); // set current colormap
  virtual void setColormapWindowCmd(char*) =0; // apply colormap to named window
};

#endif

