/* Fitsy FITS routines to read and write FITS headers.
 */

#include <xos.h>
#include <xfile.h>
#include "fitsy.h"


static int GetWordType(fptr, tbuf)
     char **fptr;
     char *tbuf;
{
  int alpha, digit, other, got;
  char *f, *t;

  /* start out clean */
  *tbuf = '\0';

  /* gather up characters until we have a delim */
  for(got=0, alpha=0, digit=0, other=0, f=*fptr, *tbuf='\0', t=tbuf;
      (*f!='\0') && (*f!=',') && (*f!=']');
      got++){
    if( isdigit((unsigned int) *f) )
      digit++;
    else if( isalpha((unsigned int) *f) )
      alpha++;
    else
      other++;
    *t++ = *f++;
  }
  /* null terminate working string */
  *t = '\0';

  /* update pointer */
  *fptr = f;

  /* determine what sort of beast we have and return */
  if( got == 0 )
    return(0);
  else if( other >0 )
    return(3);
  else if( (digit>0) && (alpha==0) )
    return(2);
  else 
    return(1);
}


/*
 *
 * ft_parsefilename -- break a string into name, extension, index, tail
 * The possibilities are:
 *
 * name[extn], name[extn,index], name[index]
 * name[extn,tail], name[extn,index,tail], name[index,tail]
 *
 * where:
 * extn is composed of alphanumerics, index is numeric, tail is anything
 * and commas are separators.
 *
 */
void ft_parsefilename(filename, name, extn, indx, tail, maxlen)
     char *filename;
     char *name;
     char *extn;
     int *indx;
     char *tail;
     int maxlen;
{
  char *f, *t;
  char *tbuf, *tbuf2;

  /* start out clean */
  *extn = '\0';
  *indx = -1;
  *tail = '\0';

  /* make sure we have something to work with */
  if( (filename == NULL) || (*filename == '\0') )
    return;

  /* basic work buffers */
  tbuf = (char *)malloc(strlen(filename)+1);
  tbuf2 = (char *)malloc(strlen(filename)+1);

  /* this will be our pointer into the filename buffer */
  f = filename;

  /* maxlen generally is 1 more than we can handle */
  maxlen--;

  /* gather up the name */
  for(*tbuf='\0', t=tbuf; (*f!='\0') && (*f!='['); )
    *t++ = *f++;
  /* null terminate working string */
  *t = '\0';
  /* copy max from the work buffer to the name buffer */
  strncpy(name, tbuf, maxlen);
  name[maxlen] = '\0';
  /* check for done */
  if( *f == '\0' )
    goto cleanup;
  /* or skip past '[' delim following the name */
  else
    f++;

  /* get token and process type */
  switch(GetWordType(&f, tbuf)){
  /* EOF */
  case 0:
    break;
  /* extn */
  case 1:
    strcpy(extn, tbuf);
    *tbuf = '\0';
    if( *f == ',' ){
      f++;
      switch(GetWordType(&f, tbuf)){
      case 2:
	*indx = strtol(tbuf, NULL, 10);
	*tbuf = '\0';
	if( *f == ',' )
	  f++;
	break;
      default:
	break;
      }
    }
    break;
  /* index */
  case 2:
    *indx = strtol(tbuf, NULL, 10);
    *tbuf = '\0';
    if( *f == ',' )
      f++;
    break;
  /* tail */
  case 3:
    break;
  }

  /* at this point, we are processing tail */
  strcpy(tbuf2, "[");
  if( *tbuf != '\0' )
    strcat(tbuf2, tbuf);
  if( *f != '\0' )
    strcat(tbuf2, f);
  strncpy(tail, tbuf2, maxlen);
  tail[maxlen] = '\0';
  /* don't bother passing back null brackets */
  if( !strcmp(tail, "[") || !strcmp(tail, "[]") || !strcmp(tail, "[][]") )
    *tail = '\0';

  /* clean up */
cleanup:
  if( tbuf != NULL )
    free(tbuf);
  if( tbuf2 != NULL )
    free(tbuf2);
}

/* Read a FITS header from the named file.
 */
File ft_fitsheadopen(filename, header, tail, len)
     char 	  *filename;	/* FITS image file name.		*/
     FITSHead 	  *header;
     char	  *tail;
     int 	  len;
{
	FITSHead  prim = NULL;
	FITSHead  head = NULL;
	File file;

	char	name[FT_FILENAME];
	char	extn[FT_FILENAME];
	int	indx = 0;

	char	*e;
	char	*i;
	char tbuf[FT_CARDLEN];

    ft_parsefilename(filename, name, extn, &indx, tail, len);

    if( !strcmp(name, "stdin")  ||
        !strcmp(name, "STDIN")  ||
        !strcmp(name, "-")      )
        file = stdin;
    else if ( (file = Open(name, "r"))   == NULL )
	return NULL;

    /* make sure this is a FITS file by looking at the first card */
    if( (Read(file, tbuf, FT_CARDLEN, 1) != 1) ||
        (strncmp(tbuf, "SIMPLE  =", 9) && strncmp(tbuf, "XTENSION=", 9)) ){
      Close(file);
      return(NULL);
    }

    if ( (prim = ft_headread0(file, tbuf, FT_CARDLEN)) == NULL ){
        Close(file);
	return NULL;
    }

    /* 	Only look in the environment if the primary header
	has no data and the user hasn't specified and thing
	in the brackets.
    */
    if ( ft_databytes(prim) == 0 && !*extn && indx == -1 ) {
	if ( (e = (char *)getenv("FITS_EXTNAME")) ) {
	    strncpy(extn, e, len);
	    extn[len-1] = '\0';
	}
	if ( (i = (char *)getenv("FITS_EXTNUM")) ) {
	    indx = atoi(i);
	}
    }
	
    head = prim;
    if ( *extn ) {		/* Extension and possible an index	*/
	do {
	  char *ext=NULL;
	  char *extcopy=(char *)strdup(extn);
	  for(ext=(char *)strtok(extcopy," :;,");
	      ext!=NULL;
	      ext=(char *)strtok(NULL," :;,")){
	    if ( ft_name(head) && *ft_name(head) && 
		 !strcasecmp(ext, ft_name(head)) &&
		 (indx == -1 || indx == ft_extn(head)) )
	      break;
	  }
	  if( extcopy )
	    free(extcopy);
	  /* if we matched an extension, we are done */
	  if( ext )
	    break;

	  ft_dataskip(file, head, NULL);

	  if ( head != prim ) ft_headfree(head, 1);
	} while ( (head = ft_headread(file)) );

	if ( !head ) {
	    ft_headfree(prim, 1);
	    Close(file);
	    return NULL;
	}
    } else if ( indx > 0 ) {	/* Extension index, not == 0		*/
	    int	i;

	for ( i = 1; i <= indx; i++ ) {
	    if ( (head = ft_headread(file)) == NULL ) {
		ft_headfree(prim, 1);
		Close(file);
		return NULL;
	    }

	    if( i != indx ){
	      ft_dataskip(file, head, NULL);
	      ft_headfree(head, 1);
	    }
	}
    } 

    if ( head != prim )
	ft_primlink(prim, head);	/* Link the primary to the header */

    if( header != NULL )
      *header = head;
    return file;
}

