#include <strstream.h>

#include "circle.h"
#include "frame.h"
#include "util.h"

Circle::Circle(Frame* p, const Vector& ctr, double r, 
	       const char* clr, const char* f, const char* t, 
	       unsigned short prop)
  : AnnuliMarker(p, ctr, 0, clr, f, t, prop)
{
  radius = r;
  strcpy(type, "circle");
  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

void Circle::updateBBox()
{
  // bound marker

  Vector c = center * parent->refToCanvas;
  Vector r = Vector(radius,-radius) * maxAnnuli();
  if (!(properties & FIXED))
    r *= parent->getZoom();
  bbox = BBox(c-r, c+r);

  // generate handles

  handle[0] = bbox.ll;
  handle[1] = bbox.lr();
  handle[2] = bbox.ur;
  handle[3] = bbox.ul();

  // bound handles

  for (int i=0; i<numHandle; i++)
    bbox.bound(handle[i]);

  // make room for handles

  bbox.expand(3);

  // calculate overall bbox

  calcAllBBox();
}

void Circle::updateCoords(Matrix& m)
{
  radius *= m[0][0];
  Marker::updateCoords(m);
}

void Circle::draw()
{
  Marker::draw();
  includeExclude();
  render(parent->pixmap, parent->refToWidget, gc, parent->getZoom());
}

void Circle::drawXOR()
{
  render(Tk_WindowId(parent->tkwin), parent->refToWindow, 
	 gcxor, parent->getZoom());
}

void Circle::drawMagnifier(const Matrix& m)
{
  XSetForeground(display, gc, color);
  includeExclude();
  render(parent->magnifierPixmap, m, gc, parent->getMagnifierZoom());
}

void Circle::edit(const Vector& v, int h)
{
  Vector r = (v-center).abs();
  radius = r[0] > r[1] ? r[0] : r[1];

  updateBBox();
  doCallBacks(&editCB);
}

void Circle::ps(int mode)
{
  Marker::ps(mode);
  psIncludeExclude();

  Vector c = center*parent->refToCanvas;

  char buf[256];
  for (int i=0; i<annuli; i++) {
    double r = radius * ratio[i];
    if (!(properties & FIXED))
      r *= parent->getZoom();

    ostrstream str(buf,256);
    str << "newpath " << c.TkCanvasPs(parent->canvas) << r
	<< " 0 360 arc stroke" << endl << ends;
    Tcl_AppendResult(parent->interp, buf, NULL);
  }
}

int Circle::isIn(const Vector& v)
{
  // v is in canvas coords

  double r = radius * maxAnnuli();
  if (properties & FIXED)
    r /= parent->getZoom();

  Vector p = v * parent->canvasToRef * Translate(-center);

  if ((p[0]*p[0] + p[1]*p[1])/(r*r) <= 1)
    return 1;
  else
    return 0;
}

void Circle::setRadius(double r)
{
  radius = r;
  updateBBox();

  doCallBacks(&editCB);
}

// private

void Circle::render(Drawable drawable, const Matrix& matrix, 
		    GC gc, double zoom)
{
  for (int i=0; i<annuli; i++) {
    Vector c = center * matrix;
    Vector r = Vector(radius, radius) * ratio[i];

    if (!(properties & FIXED))
      r *= zoom;

    Vector ll = (c-r).round();
    Vector size = (r*2).round();

    // Verify size is positive
    // XDrawArc is sensative to bad data, and may hang the XServer

    if (size[0]>0 && size[1]>0)
      XDrawArc(display, drawable, gc, (int)ll[0], (int)ll[1],
	       (int)size[0], (int)size[1], 0, 360*64);
  }
}

// list

void Circle::list(ostream& str, int sys, int format)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector r = parent->mapDistFromRef(Vector(radius,radius),
					(Frame::CoordSystem)sys,Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << setprecision(8) << v  << r[0];
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::ALTAZ:
  case Frame::LINEAR:
  case Frame::ICRS:
    Vector r = parent->mapDistFromRef(Vector(radius,radius),
				      (Frame::CoordSystem)sys,Frame::ARCSEC);
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << setprecision(8) << v << r[0] << '"';
      }
      break;
    case Frame::HMS:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << ' ' << ra << ' ' << dec << ' ' << r[0] << '"';
      }
      break;
    }
    break;
  }

  listProperties(str);
}

void Circle::listSAOtng(ostream& str, int sys, int format)
{
  if (text[0])
    str << '#' << text << endl;

  if (properties&INCLUDE)
    str << '+';
  else
    str << '-';

  // radius is always in image coords

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] 
	  << ',' << radius << ") ";
    }
    break;

  case Frame::PHYSICAL:
    {
      Vector r = parent->mapDistFromRef(Vector(radius,radius),
					(Frame::CoordSystem)sys,Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] 
	  << ',' << r[0] << ") ";
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::ALTAZ:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	    << radius << ") ";
      }
      break;
    case Frame::HMS:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',' << radius << ") ";
      }
      break;
    }
    break;
  }

  str << "# " << colorName;
}

void Circle::listPros(ostream& str, int sys, int format)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << type << ' ' << setprecision(8) << v << radius;
    }
    break;

  case Frame::PHYSICAL:
    {
      str << "physical; ";
      Vector r = parent->mapDistFromRef(Vector(radius,radius),
					(Frame::CoordSystem)sys,Frame::PIXELS);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << ' ' << setprecision(8) << v << r[0];
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::ALTAZ:
  case Frame::LINEAR:
  case Frame::ICRS:
    parent->listProsCoordSystem(str, (Frame::CoordSystem)sys);

    // display wcs radius in ARCSEC
    Vector r = parent->mapDistFromRef(Vector(radius,radius),
				      (Frame::CoordSystem)sys,Frame::ARCSEC);

    str << "; ";
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << ' ' << setprecision(8) << v[0] << "d " << v[1] << "d " 
	    << r[0] << '"';
      }
      break;
    case Frame::HMS:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	if (dec[0]=='+')
	  str << type << ' ' << ra << ' ' << dec+1 << ' ' << r[0] << '"';
	else
	  str << type << ' ' << ra << ' ' << dec << ' ' << r[0] << '"';
      }
      break;
    }
    break;
  }
}

void Circle::listSAOimage(ostream& str)
{
  // all coords are in image coords

  if (!(properties&INCLUDE))
    str << '-';

  Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
  str << "circle(" << setprecision(8) << v[0] << ',' << v[1] 
      << ',' << radius << ") ";
}
