#ifndef __fitsdata_h__
#define __fitsdata_h__

#include <iostream.h>
#include <strstream.h>

#include "vector.h"
#include "frame.h"
#include "file.h"

class ColorScale;
class FitsFile;

class FitsData {
protected:
  int width;
  int height;
  long size;
  char buf[32];

  int byteswap;

  double bscale;
  double bzero;
  int hasScaling;

  int hasBlank;
  int blank;
  float nanf;          // generated NaN
  double nand;         // generated NaN

  double high;         // bscale applied
  double low;          // bscale applied

  float zHigh;         // bscale applied
  float zLow;          // bscale applied
  double uLow;	       // bscale applied		    
  double uHigh;	       // bscale applied		    

  int scanValid;

  int incr;
  int sampleValid;

  float zContrast;
  int zSampleSize;
  int zSampleLine;
  int zscaleValid;

  Frame::ClipMode clipMode;
  Frame::MinMaxMode mmMode;

  double datamin;      // bscale applied
  double datamax;      // bscale applied
  int hasdatamin;

  double irafmin;      // bscale applied
  double irafmax;      // bscale applied
  int hasirafmin;

  void swap2(char* src, char* dest) {
    *(dest  ) = *(src+1);
    *(dest+1) = *(src  );
  }
  void swap4(char* src, char* dest) {
    *(dest  ) = *(src+3);
    *(dest+1) = *(src+2);
    *(dest+2) = *(src+1);
    *(dest+3) = *(src  );
  }
  void swap8(char* src, char* dest) {
    *(dest  ) = *(src+7);
    *(dest+1) = *(src+6);
    *(dest+2) = *(src+5);
    *(dest+3) = *(src+4);
    *(dest+4) = *(src+3);
    *(dest+5) = *(src+2);
    *(dest+6) = *(src+1);
    *(dest+7) = *(src  );
  }

  // zscale constants

  enum PixelType {GOOD_PIXEL, BAD_PIXEL, REJECT_PIXEL};
  static const int MIN_NPIXELS = 5;	   // smallest permissible sample
  static const float MAX_REJECT = 0.5;    // max frac. of pixels to be rejected
  static const float KREJ = 2.5;          // k-sigma pixel rejection factor
  static const int MAX_ITERATIONS = 5;    // maximum number of fitline iterations
  static const int INDEF = 0;

  int zSubSample(float*, float*, int, int);
  int zFitLine(float*, int, float*, float*, float, int, int);
  void zFlattenData (float*, float*, float*, int npix, float, float dz);
  int zComputeSigma (float*, short*, int npix, float*, float*);
  int zRejectPixels (float*, float*, float*, short*, int, double*, double*, double*, 
		     double*, float, int);

public:
  FitsData(FitsFile*);
  virtual ~FitsData();

  virtual const char* getValue(const Vector&) =0;
  virtual float getValueFloat(const Vector&) =0;
  virtual float getValueFloat(int) =0;
  virtual double getValueDouble(const Vector&) =0;
  virtual double getValueDouble(int) =0;

  virtual const char* getMin() =0;
  virtual const char* getMax() =0;
  virtual double getMinDouble() =0;
  virtual double getMaxDouble() =0;

  const char* getLow();
  const char* getHigh();
  double getLowDouble() {return low;}
  double getHighDouble() {return high;}

  void setClip(double l, double h) {low = l; high = h;}

  virtual void updateMinMax(Frame::MinMaxMode,int) =0;
  virtual void updateClip(Frame::ClipMode,Frame::MinMaxMode,
			  float,int,int,int,double,double,int) =0;

  int hasDATAMIN() {return hasdatamin;}
  int hasIRAFMIN() {return hasirafmin;}

  virtual void bin(double*,int,double,double) =0;
};

template<class T>
class FitsDatam : public FitsData {
private:
  T* data;
  T min;               // bscale not applied
  T max;               // bscale not applied
  T mins;              // bscale not applied
  T maxs;              // bscale not applied

  T minX;              // x coord of min value, bscale not applied
  T minY;              // y coord of min value, bscale not applied
  T maxX;              // x coord of max value, bscale not applied
  T maxY;              // y coord of max value, bscale not applied

  T Tmin();
  T Tmax();

  void scan();
  void sample();
  void output(ostrstream&, T);

  void zscale();
  int zSampleImage(float**);

public:
  FitsDatam(FitsFile*);

  T swap(T*);

  const char* getValue(const Vector&);
  float getValueFloat(const Vector&);
  float getValueFloat(int i)
  {
    // no bounds checking, we need the speed

    // first, try the simple case
    if (!byteswap && !hasBlank && !hasScaling) return data[i];

    register T value = !byteswap ? data[i] : swap(data+i);
    if (hasBlank && value == blank)
      return nand;

    return hasScaling ? value * bscale + bzero : value;
  }

  double getValueDouble(const Vector&);
  double getValueDouble(int i)
  {
    // no bounds checking, we need the speed

    // first, try the simple case
    if (!byteswap && !hasBlank && !hasScaling) return data[i];

    register T value = !byteswap ? data[i] : swap(data+i);
    if (hasBlank && value == blank)
      return nand;

    return hasScaling ? value * bscale + bzero : value;
  }

  const char* getMin();
  const char* getMax();
  double getMinDouble();
  double getMaxDouble();

  void updateMinMax(Frame::MinMaxMode,int);
  void updateClip(Frame::ClipMode,Frame::MinMaxMode,float,int,int,int,double,double,int);
  void bin(double*,int,double,double);
};

#endif
