#ifndef __fitsimage_h__
#define __fitsimage_h__

#include "frame.h"
#include "vector.h"
#include "fitsdata.h"

class FitsFile;

extern "C" {
  #include "wcs.h"
}

static char *wcsName[] = { // wcs coordinate system strings (for use with wcssubs)
  "fk4",
  "fk5",
  "galactic",
  "ecliptic",
  "altaz",
  "linear",
  "icrs"
};

class FitsImage {
protected:
  enum Gravity {NOGRAVITY,GRAVITY};

  Frame* parent;            // pointer to parent
  int valid;                // is load a success
  char* objectName;         // object name
  char* fileName;           // base name of image
  char* fullFileName;       // full name of image
  FitsFile* fits;           // native fits ext
  FitsFile* image;          // fits image
  FitsData* data;           // raw fits image data

  int isHist_;
  Vector histCursor;        // point image is binned about

  int width_;
  int height_;
  Vector center;

  double rotation;          // local rotation
  Vector zoom;              // local scaling;
  Frame::Orientation orientation; // local orientation
  Vector origin;            // local translation in image space

  static Matrix baseDetectorToData; // base Scale from Detector to Data coords for mosaic

  int keyLTMV;
  int keyATMV;
  int keyDTMV;

  int keySEC;
  int keySIZE;

  Matrix amplifierToCanvas;
  Matrix amplifierToPhysical;
  Matrix amplifierToRef;
  Matrix amplifierToPanner;
  Matrix amplifierToUser;
  Matrix canvasToAmplifier;
  Matrix canvasToPhysical;
  Matrix canvasToData;
  Matrix canvasToDetector;
  Matrix canvasToImage;
  Matrix physicalToAmplifier;
  Matrix physicalToCanvas;
  Matrix physicalToData;
  Matrix physicalToDetector;
  Matrix physicalToImage;
  Matrix physicalToRef;
  Matrix physicalToUser;
  Matrix physicalToPanner;
  Matrix dataToCanvas;
  Matrix dataToPhysical;
  Matrix dataToImage;
  Matrix dataToMagnifier;
  Matrix dataToRef;
  Matrix dataToPanner;
  Matrix dataToUser;
  Matrix dataToWidget;
  Matrix detectorToCanvas;
  Matrix detectorToPhysical;
  Matrix detectorToPanner;
  Matrix detectorToRef;
  Matrix detectorToUser;
  Matrix imageToCanvas;
  Matrix imageToPhysical;
  Matrix imageToData;
  Matrix imageToRef;
  Matrix imageToPanner;
  Matrix imageToUser;
  Matrix magnifierToData;
  Matrix pannerToAmplifier;
  Matrix pannerToPhysical;
  Matrix pannerToData;
  Matrix pannerToDetector;
  Matrix pannerToImage;
  Matrix refToPhysical;
  Matrix refToData;
  Matrix refToImage;
  Matrix refToAmplifier;
  Matrix refToDetector;
  Matrix userToAmplifier;
  Matrix userToPhysical;
  Matrix userToData;
  Matrix userToDetector;
  Matrix userToImage;
  Matrix widgetToData;

  Matrix binToData;
  Matrix dataToBin;

  BBox detsize;             // mosaic size
  int mosaicMode;           // flag to indicate if part of mosaic
  int firstMosaic;          // flag to indicate if first mosaic

  WorldCoor* wcs;           // wcs

  FitsImage* next_;          // next in list

  void process();
  void processMosaic(const char*);
  void load();
  void processKeywords();
  Matrix nextHist(Vector,Gravity);
  Vector getHistCenter();

  char* root(const char*);
  BBox parseSection(char*);
  const char* coordSystemToWCS(Frame::CoordSystem);
  Frame::CoordSystem wcsToCoordSystem(const char*);

public:
  FitsImage(Frame*, int, int);
  virtual ~FitsImage();

  int isValid() {return valid;}
  FitsFile* getFitsFile() {return fits;}

  int isHist() {return isHist_;}
  int isImage() {return fits ? fits->isImage() : 0;}
  int isTable() {return fits ? fits->isTable() : 0;}
  void initHist();
  Matrix updateHist();
  Matrix updateHist(Vector);
  Matrix updateHistUser(Vector);
  Vector getHistSize();
  Vector getHistCursor() {return histCursor;}

  FitsImage* next() {return next_;}
  void setNext(FitsImage* n) {next_ = n;}
  const char* getObjectName() {return objectName;}
  const char* getFileName() {return fileName;}
  const char* getFullFileName() {return fullFileName;}

  int width() {return width_;}
  int height() {return height_;}
  Vector getCenter();
  Vector getSize();

  int hasLTMV() {return keyLTMV;}
  int hasATMV() {return keyATMV;}
  int hasDTMV() {return keyDTMV;}
  int isMosaic() {return (mosaicMode && keySEC && keySIZE);}

  char* pix2wcs(Vector, Frame::CoordSystem, Frame::CoordFormat, char*, int);
  Vector pix2wcs(Vector);
  Vector pix2wcs(Vector, Frame::CoordSystem);
  Vector wcs2pix(Vector);
  Vector wcs2pix(Vector, Frame::CoordSystem);

  Vector wcs2wcs(Vector, Frame::CoordSystem, Frame::CoordSystem);

  Vector wcs2pixDist(Vector);
  Vector pix2wcsDist(Vector);
  double wcsdist(Vector, Vector);

  Frame::Orientation getWCSOrientation();
  double getWCSRotation();

  int hasWCS();
  int hasWCSEqu();
  Frame::CoordSystem getWCSCoordSystem();
  const char* getWCSSystem();
  Vector getWCScrval();
  Vector getWCScrpix();
  Vector getWCScdelt();

  void updateMatrices(Matrix&, Matrix&, Matrix&, Matrix&);
  void updateMagnifierMatrices(Matrix&);

  Matrix& getAmplifierToPhysical() {return amplifierToPhysical;}
  Matrix& getAmplifierToRef() {return amplifierToRef;}
  Matrix& getAmplifierToUser() {return amplifierToUser;}
  Matrix& getCanvasToAmplifier() {return canvasToAmplifier;}
  Matrix& getCanvasToPhysical() {return canvasToPhysical;}
  Matrix& getCanvasToData() {return canvasToData;}
  Matrix& getCanvasToDetector() {return canvasToDetector;}
  Matrix& getCanvasToImage() {return canvasToImage;}
  Matrix& getPhysicalToAmplifier() {return physicalToAmplifier;}
  Matrix& getPhysicalToDetector() {return physicalToDetector;}
  Matrix& getPhysicalToImage() {return physicalToImage;}
  Matrix& getPhysicalToRef() {return physicalToRef;}
  Matrix& getPhysicalToUser() {return physicalToUser;}
  Matrix& getDataToCanvas() {return dataToCanvas;}
  Matrix& getDataToPhysical() {return dataToPhysical;}
  Matrix& getDataToPanner() {return dataToPanner;}
  Matrix& getDataToRef() {return dataToRef;}
  Matrix& getDataToUser() {return dataToUser;}
  Matrix& getDataToWidget() {return dataToWidget;}
  Matrix& getDetectorToPhysical() {return detectorToPhysical;}
  Matrix& getDetectorToRef() {return detectorToRef;}
  Matrix& getDetectorToUser() {return detectorToUser;}
  Matrix& getImageToPhysical() {return imageToPhysical;}
  Matrix& getImageToRef() {return imageToRef;}
  Matrix& getImageToUser() {return imageToUser;}
  Matrix& getMagnifierToData() {return magnifierToData;}
  Matrix& getRefToAmplifier() {return refToAmplifier;}
  Matrix& getRefToPhysical() {return refToPhysical;}
  Matrix& getRefToData() {return refToData;}
  Matrix& getRefToDetector() {return refToDetector;}
  Matrix& getRefToImage() {return refToImage;}
  Matrix& getPannerToAmplifier() {return pannerToAmplifier;}
  Matrix& getPannerToPhysical() {return pannerToPhysical;}
  Matrix& getPannerToData() {return pannerToData;}
  Matrix& getPannerToDetector() {return pannerToDetector;}
  Matrix& getPannerToImage() {return pannerToImage;}
  Matrix& getUserToAmplifier() {return userToAmplifier;}
  Matrix& getUserToPhysical() {return userToPhysical;}
  Matrix& getUserToData() {return userToData;}
  Matrix& getUserToDetector() {return userToDetector;}
  Matrix& getUserToImage() {return userToImage;}
  Matrix& getWidgetToData() {return widgetToData;}

// we assume that 'data' is valid

  const char* getValue(const Vector& v) {return data->getValue(v);}
  float getValueFloat(const Vector& v) {return data->getValueFloat(v);}
  float getValueFloat(int i) {return data->getValueFloat(i);}
  double getValueDouble(const Vector& v) {return data->getValueDouble(v);}
  double getValueDouble(int i) {return data->getValueDouble(i);}
  void setClip(double l, double h) {data->setClip(l,h);}

  const char* getMin();
  const char* getMax();
  double getMinDouble();
  double getMaxDouble();

  const char* getLow();
  const char* getHigh();
  double getLowDouble();
  double getHighDouble();

  void updateMinMax(Frame::MinMaxMode mm, int mmi);
  void updateClip(Frame::ClipMode cm, Frame::MinMaxMode mm, float zc, int zs, int zl,
		  int mmi, double ul, double uh, int fs);

  int hasDATAMIN();
  int hasIRAFMIN();

  void bin(double* b, int l, double mn, double mx) {data->bin(b,l,mn,mx);}

  char* displayHeader();
};

class FitsImageMaster : public FitsImage
{
public:
  FitsImageMaster(Frame* p, int m, int fm) : FitsImage(p,m,fm) {}
  virtual ~FitsImageMaster();
};

class FitsImageSlave : public FitsImage
{
public:
  FitsImageSlave(Frame*, const char*, int, int, FitsFile*);
  virtual ~FitsImageSlave() {}
};

class FitsImageFitsAlloc : public FitsImageMaster
{
public:
  FitsImageFitsAlloc(Frame*, const char*, int, int);
};

class FitsImageFitsChannel : public FitsImageMaster
{
public:
  FitsImageFitsChannel(Frame*, Tcl_Interp*, const char*, const char*, int, int);
};

class FitsImageFitsMMap : public FitsImageMaster
{
public:
  FitsImageFitsMMap(Frame*, const char*, int, int);
};

class FitsImageFitsShare : public FitsImageMaster
{
public:
  FitsImageFitsShare(Frame*, int, int, const char*, int, int);
};

class FitsImageArrAlloc : public FitsImageMaster
{
public:
  FitsImageArrAlloc(Frame*, const char*, int, int);
};

class FitsImageArrChannel : public FitsImageMaster
{
public:
  FitsImageArrChannel(Frame*, Tcl_Interp*, const char*, const char*, int, int);
};

class FitsImageArrMMap : public FitsImageMaster
{
public:
  FitsImageArrMMap(Frame*, const char*, int, int);
};

class FitsImageArrShare : public FitsImageMaster
{
public:
  FitsImageArrShare(Frame*, int, int, const char*, int, int);
};

class FitsImageMosaicMMap : public FitsImageMaster
{
public:
  FitsImageMosaicMMap(Frame*, const char*);
};

class FitsImageMosaicShare : public FitsImageMaster
{
public:
  FitsImageMosaicShare(Frame*, int, int, const char*);
};

class FitsImageMosaicNextMap : public FitsImageMaster
{
public:
  FitsImageMosaicNextMap(Frame*, const char*, FitsFile*);
};

class FitsImageMosaicAlloc : public FitsImageMaster
{
public:
  FitsImageMosaicAlloc(Frame*, const char*);
};

class FitsImageMosaicChannel : public FitsImageMaster
{
public:
  FitsImageMosaicChannel(Frame*, Tcl_Interp*, const char*, const char*);
};

class FitsImageMosaicNextAlloc : public FitsImageMaster
{
public:
  FitsImageMosaicNextAlloc(Frame*, const char*, FitsFile*);
};

class FitsImageMosaicNextChannel : public FitsImageMaster
{
public:
  FitsImageMosaicNextChannel(Frame*, const char*, FitsFile*);
};

class FitsImageIIS : public FitsImageMaster
{
public:
  FitsImageIIS(Frame*, int, int);
  void iisErase();
  const char* iisGet(int x, int y);
  void iisSet(int x, int y, const char* d, int l);
  void iisWCS(const char*, const Matrix&);
};

#endif
