#include <stdlib.h>
#include <string.h>

#include <iostream.h>
#include <strstream.h>

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "mmap.h"

#include "marker.h"
#include "annuli.h"
#include "circle.h"
#include "ellipse.h"
#include "rectangle.h"
#include "polygon.h"
#include "line.h"
#include "text.h"
#include "ruler.h"
#include "point.h"

// Marker Parser Stuff

#undef yyFlexLexer
#define yyFlexLexer mkFlexLexer
#include <FlexLexer.h>

mkFlexLexer* mklexx = NULL;    // used by mklex
extern Frame* fr;              // used by mkerror
extern int mkparse(void*);

int mklex()
{
  return (mklexx ? mklexx->yylex() : 0);
}

void mkerror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = mklexx ? mklexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

// Marker Commands

void Frame::createCircleCmd(const Vector& center, CoordSystem sys,
			    double radius, CoordFormat format,
			    const char* color, const char* font,
			    const char* text, unsigned short prop)
{
  Vector r = mapDistToRef(Vector(radius,radius),sys,format);
  Marker* m = new Circle(this, map(center,sys,REF), r[0], 
			 color, font, text, prop);
  createMarker(m);
}

void Frame::createEllipseCmd(const Vector& center, CoordSystem sys,
			     const Vector& radius, CoordFormat format,
			     double angle,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Ellipse(this, map(center,sys,REF), 
			  mapDistToRef(radius,sys,format),
			  angle, color, font, text, prop);
  createMarker(m);
}

void Frame::createRectangleCmd(const Vector& center, CoordSystem sys,
			     const Vector& size, CoordFormat format,
			     double angle,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Rectangle(this, map(center,sys,REF), 
			    mapDistToRef(size,sys,format),
			    angle, color, font, text, prop);
  createMarker(m);
}

void Frame::createPolygonCmd(const Vector& center, const Vector& bb, 
			     CoordSystem sys,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Polygon(this, map(center,sys,REF), bb,
			  color, font, text, prop);
  createMarker(m);
}

void Frame::createPolygonCmd(List<Vertex>* list, CoordSystem sys,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  list->head();
  do
    list->current()->vector = map(list->current()->vector, sys, REF);
  while (list->next());

  Marker* m = new Polygon(this, list, color, font, text, prop);
  createMarker(m);
}

void Frame::createLineCmd(const Vector& center, const Vector& p2, 
			  CoordSystem sys,
			  int arrow1, int arrow2,
			  const char* color, const char* font,
			  const char* text, unsigned short prop)
{
  Marker* m = new Line(this, map(center,sys,REF), map(p2,sys,REF), 
		       arrow1, arrow2,
		       color, font, text, prop);
  createMarker(m);
}

void Frame::createTextCmd(const Vector& center, CoordSystem sys,
				 const char* color, const char* font,
				 const char* text, unsigned short prop)
{
  Marker* m = new Text(this, map(center,sys,REF), color, font, text, prop);
  createMarker(m);
}

void Frame::createRulerCmd(const Vector& center, const Vector& p2,
			   CoordSystem sys, CoordFormat format,
			   const char* color, const char* font,
			   const char* text, unsigned short prop)
{
  Marker* m = new Ruler(this, map(center,sys,REF), map(p2,sys,REF), format,
			color, font, text, prop);
  createMarker(m);
}

void Frame::createCirclePointCmd(const Vector& center, CoordSystem sys,
				 const char* color, const char* font,
				 const char* text, unsigned short prop)
{
  Marker* m = new CirclePoint(this, map(center,sys,REF), 
			      color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxPointCmd(const Vector& center, CoordSystem sys,
			      const char* color, const char* font,
			      const char* text, unsigned short prop)
{
  Marker* m = new BoxPoint(this, map(center,sys,REF), color, font, text, prop);
  createMarker(m);
}

void Frame::createDiamondPointCmd(const Vector& center, CoordSystem sys,
				  const char* color, const char* font,
				  const char* text, unsigned short prop)
{
  Marker* m = new DiamondPoint(this, map(center,sys,REF), 
			       color, font, text, prop);
  createMarker(m);
}

void Frame::createCrossPointCmd(const Vector& center, CoordSystem sys,
				const char* color, const char* font,
				const char* text, unsigned short prop)
{
  Marker* m = new CrossPoint(this, map(center,sys,REF), 
			     color, font, text, prop);
  createMarker(m);
}

void Frame::createExPointCmd(const Vector& center, CoordSystem sys,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new ExPoint(this, map(center,sys,REF), color, font, text, prop);
  createMarker(m);
}

void Frame::createArrowPointCmd(const Vector& center, CoordSystem sys,
				const char* color, const char* font,
				const char* text, unsigned short prop)
{
  Marker* m = new ArrowPoint(this, map(center,sys,REF), 
			     color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxCirclePointCmd(const Vector& center, CoordSystem sys,
				    const char* color, const char* font,
				    const char* text, unsigned short prop)
{
  Marker* m = new BoxCirclePoint(this, map(center,sys,REF), 
				 color, font, text, prop);
  createMarker(m);
}

void Frame::createMarker(Marker* m)
{
  markers.append(m);

  // now update new marker

  update(PIXMAP, m->getAllBBox());

  // and return id

  printIntegerTcl(m->getId());
}

void Frame::getMarkerCenterCmd(int id, CoordSystem sys)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	printCoordTcl(markers.current()->getCenter(), REF, sys, DEFAULT);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerCenterCmd(int id, CoordSystem wcs, CoordFormat format)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	getWCSCoordCmd(markers.current()->getCenter(), REF, wcs, format);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerCircleRadiusCmd(int id, CoordSystem sys, 
				     CoordFormat format)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	double r = ((Circle*)(markers.current()))->getRadius();
	Vector v = mapDistFromRef(Vector(r,0), sys, format);
	printDoubleTcl(v[0], DEFAULT);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerColorCmd()
{
  // return first found

  if (markers.head())
    do
      if (markers.current()->isSelected()) {
	Tcl_AppendResult(interp,  markers.current()->getColorName(), NULL);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerColorCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	Tcl_AppendResult(interp,  markers.current()->getColorName(), NULL);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerEllipseRadiusCmd(int id, CoordSystem sys, 
				      CoordFormat format)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	Vector r = ((Ellipse*)(markers.current()))->getRadius();
	printVectorTcl(mapDistFromRef(r,sys,format),DEFAULT);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerEllipseAngleCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	printDoubleTcl(radToDeg(((Ellipse*)(markers.current()))->getAngle()),
		       DEFAULT);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerFontCmd()
{
  if (markers.head())
    do
      if (markers.current()->isSelected()) {
	Tcl_AppendResult(interp, markers.current()->getFont(), NULL);
	return;
      }
    while (markers.next());
}

void Frame::getMarkerFontCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	Tcl_AppendResult(interp, markers.current()->getFont(), NULL);
	return;
      }
    while (markers.next());
}

void Frame::getMarkerHandleCmd(const Vector& v)
{
  if (markers.head())
    do
      if (markers.current()->isSelected()) {
	int h = markers.current()->onHandle(v);
	if (h) {
	  char buf [16];
	  ostrstream str(buf,16);
	  str << markers.current()->getId() << ' ' << h << ends;
	  Tcl_AppendResult(interp, buf, NULL);
	  return;
	}
      }
    while (markers.next());

    Tcl_AppendResult(interp, "0 0", NULL);
}

void Frame::getMarkerIdCmd(const Vector& v)
{
  // v is in canvas coords

  if (markers.head())
    do
      if (markers.current()->isIn(v)) {
	printIntegerTcl(markers.current()->getId());
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerLineCmd(int id, CoordSystem sys)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	printCoordTcl(((Line*)markers.current())->getP1(), REF, sys, DEFAULT);
	printCoordTcl(((Line*)markers.current())->getP2(), REF, sys, DEFAULT);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerLineCmd(int id, CoordSystem wcs, CoordFormat format)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	getWCSCoordCmd(((Line*)markers.current())->getP1(), REF, wcs, format);
	Tcl_AppendResult(interp, " ", NULL);
	getWCSCoordCmd(((Line*)markers.current())->getP2(), REF, wcs, format);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerLineArrowCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	if (((Line*)markers.current())->getP1Arrow())
	  Tcl_AppendResult(interp, "1", NULL);
	else
	  Tcl_AppendResult(interp, "0", NULL);

	if (((Line*)markers.current())->getP2Arrow())
	  Tcl_AppendResult(interp, " 1", NULL);
	else
	  Tcl_AppendResult(interp, " 0", NULL);

	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerPolygonAngleCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	printDoubleTcl(radToDeg(((Polygon*)(markers.current()))->getAngle()), 
		       DEFAULT);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerPolygonSegmentCmd(const Vector& v)
{
  if (markers.head())
    do
      if (markers.current()->isSelected() && markers.current()->isPolygon()) {
	int s = ((Polygon*)(markers.current()))->getSegment(v);
	if (s) {
	  char buf [16];
	  ostrstream str(buf,16);
	  str << markers.current()->getId() << ' ' << s << ends;
	  Tcl_AppendResult(interp, buf, NULL);
	  return;
	}
      }
    while (markers.next());

    Tcl_AppendResult(interp, "0 0", NULL);
}


void Frame::getMarkerPropertyCmd(unsigned short which)
{
  // return first selected found

  if (markers.head())
    do
      if (markers.current()->isSelected()) {
	markers.current()->getProperty(which) ?
	  Tcl_AppendResult(interp, "1", NULL) : Tcl_AppendResult(interp, "0", 
								 NULL);
	return;
      }
    while (markers.next());

  // else, return 'off'

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerPropertyCmd(int id, unsigned short which)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	markers.current()->getProperty(which) ?
	  Tcl_AppendResult(interp, "1", NULL) : Tcl_AppendResult(interp, "0",
								 NULL);
	return;
      }
    while (markers.next());

  // else, return 'off'

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerRectangleSizeCmd(int id, CoordSystem sys, 
				      CoordFormat format)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	Vector s = ((Rectangle*)(markers.current()))->getSize();
	printVectorTcl(mapDistFromRef(s,sys,format),DEFAULT);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerRectangleAngleCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	printDoubleTcl(radToDeg(((Rectangle*)(markers.current()))->getAngle()),
		       DEFAULT);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerRulerDistanceCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	printDoubleTcl(((Ruler*)(markers.current()))->getDistance(),
		       DEFAULT);
	printVectorTcl(((Ruler*)(markers.current()))->getAxisDistance(),
		       DEFAULT);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerRulerPointCmd(int id, CoordSystem sys)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	printCoordTcl(((Ruler*)markers.current())->getP1(), REF, sys, DEFAULT);
	printCoordTcl(((Ruler*)markers.current())->getP2(), REF, sys, DEFAULT);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerRulerPointCmd(int id, CoordSystem wcs, CoordFormat format)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	getWCSCoordCmd(((Ruler*)markers.current())->getP1(), REF, wcs, format);
	Tcl_AppendResult(interp, " ", NULL);
	getWCSCoordCmd(((Ruler*)markers.current())->getP2(), REF, wcs, format);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerSelectedCmd()
{
  char buf[8];
  if (markers.head())
    do
      if (markers.current()->isSelected()) {
	ostrstream str(buf,8);
	str << markers.current()->getId() << ' ' << ends;
	Tcl_AppendResult(interp, buf, NULL);
      }
    while (markers.next());
}

void Frame::getMarkerSelectedCmd(const Vector& v)
{
  char buf[8];
  if (markers.head())
    do
      if (markers.current()->isIn(v) && markers.current()->isSelected()) {
	ostrstream str(buf,8);
	str << markers.current()->getId() << ' ' << ends;
	Tcl_AppendResult(interp, buf, NULL);
	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerSelectedNumberCmd()
{
  int count=0;
  if (markers.head())
    do
      if (markers.current()->isSelected())
	count++;
    while (markers.next());

  printIntegerTcl(count);
}

void Frame::getMarkerTextCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	Tcl_AppendResult(interp, markers.current()->getText(), NULL);
	return;
      }
    while (markers.next());
}

void Frame::getMarkerTypeCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	Tcl_AppendResult(interp, markers.current()->getType(), NULL);
	return;
      }
    while (markers.next());
}

void Frame::markerBackCmd()
{
  Marker* m = markers.tail();
  while (m) {
    if (m->isSelected()) {
      Marker* prev = markers.extractPrev(m);
      markers.append(m);
      update(PIXMAP, m->getAllBBox());
      m = prev;
    }
    else
      m = m->previous();
  }
}

void Frame::markerBackCmd(int id)
{
  Marker* m = markers.tail();
  while (m) {
    if (m->getId() == id) {
      markers.extractPrev(m);
      markers.append(m);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    else
      m = m->previous();
  }
}

void Frame::markerCallBackCmd(int id, Marker::Callback cb, const char* p, 
			      const char* a)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	result = markers.current()->setCallBack(cb, p, a);
	return;
      }

    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerCircleRadiusCmd(int id, double r, CoordSystem sys, 
				  CoordFormat format)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// it may shrink
	update(PIXMAP, markers.current()->getAllBBox());
	Vector rr = mapDistToRef(Vector(r,0),sys,format);
	((Circle*)(markers.current()))->setRadius(rr[0]);
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerColorCmd(const char* clr)
{
  if (markers.head())
    do
      if (markers.current()->isSelected()) {
	markers.current()->setColor(clr);
	update(PIXMAP, markers.current()->getAllBBox());
      }
    while (markers.next());
}

void Frame::markerColorCmd(int id, const char* clr)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	markers.current()->setColor(clr);
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());
}

void Frame::markerCommandCmd(const char* cmd)
{
  char buf[1024];
  ostrstream ostr(buf,1024);
  ostr << cmd << endl << ends; // lets terminate it in case

  istrstream istr(buf,1024);
  parseMarker(istr);
}
 
void Frame::markerDeleteCallBackCmd(int id, Marker::Callback cb, const char* p)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	result = markers.current()->deleteCallBack(cb, p);
	return;
      }

    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerDeleteCmd()
{
  Marker* m = markers.head();
  while (m) {
    if (m->isSelected() && m->canDelete()) {
      Marker* next = markers.extractNext(m);
      update(PIXMAP, m->getAllBBox());
      delete m;
      m = next;
    }
    else
      m = m->next();
  }
}

void Frame::markerDeleteCmd(int id)
{
  Marker* m = markers.head();
  while (m) {
    if (m->getId() == id) {
      markers.extractNext(m);
      update(PIXMAP, m->getAllBBox());
      delete m;
      return;
    }
    else
      m = m->next();
  }
}

void Frame::markerDeleteAllCmd()
{
  Marker* m = markers.head();
  while (m) {
    if (m->canDelete()) {
      update(PIXMAP, m->getAllBBox());
      Marker* next = markers.extractNext(m);
      delete m;
      m = next;
    }
    else
      m = m->next();
  }
}

void Frame::markerEditBeginCmd(int id, int h)
{
  // remember which marker is being edited

  if (markers.head())
    do
      if (markers.current()->getId() == id && markers.current()->canEdit()) {
	editMarker = markers.current();
	editMarker->editBegin(h);
	return;
      }
    while (markers.next());

  editMarker = NULL;
}

void Frame::markerEditBeginCmd(const Vector& v, int h)
{
  // remember which marker is being edited

  if (markers.head())
    do
      if (markers.current()->isSelected() && markers.current()->canEdit()) {
	editMarker = markers.current();
	editMarker->editBegin(h);
	return;
      }
    while (markers.next());

  editMarker = NULL;
}

void Frame::markerEditMotionCmd(const Vector& v, int h)
{
  if (editMarker) {

    // erase current marker now

    redraw(editMarker->getBBox());
    forceUpdate();

    editMarker->edit(v*canvasToRef, h);
    editMarker->drawXOR();
  }
}

void Frame::markerEditEndCmd()
{
  if (editMarker)
    editMarker->editEnd();
  editMarker = NULL;

  update(PIXMAP);
}

void Frame::markerEllipseRadiusCmd(int id, const Vector& radius, 
				   CoordSystem sys, CoordFormat format)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// it may shrink
	update(PIXMAP, markers.current()->getAllBBox());
	((Ellipse*)(markers.current()))->setRadius(mapDistToRef(radius,
								sys,format));
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerEllipseAngleCmd(int id, double angle)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// it may shrink
	update(PIXMAP, markers.current()->getAllBBox());
	((Ellipse*)(markers.current()))->setAngle(angle);
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerFontCmd(const char* f)
{
  if (markers.head())
    do
      if (markers.current()->isSelected()) {
	// things can shrink, so do before and after
	update(PIXMAP, markers.current()->getAllBBox());
	markers.current()->setFont(f);
	update(PIXMAP, markers.current()->getAllBBox());
      }
    while (markers.next());
}

void Frame::markerFontCmd(int id, const char* f)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// things can shrink, so do before and after
	update(PIXMAP, markers.current()->getAllBBox());
	markers.current()->setFont(f);
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());
}

void Frame::markerFrontCmd()
{
  Marker* m = markers.head();
  while (m) {
    if (m->isSelected()) {
      Marker* next = markers.extractNext(m);
      markers.insertHead(m);
      update(PIXMAP, m->getAllBBox());
      m = next;
    }
    else
      m = m->next();
  }
}

void Frame::markerFrontCmd(int id)
{
  Marker* m = markers.head();
  while (m) {
    if (m->getId() == id) {
      markers.extractNext(m);
      markers.insertHead(m);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    else
      m = m->next();
  }
}

void Frame::markerLineCmd(int id, const Vector& p1, const Vector& p2, 
			  CoordSystem sys)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// it may shrink
	update(PIXMAP, markers.current()->getAllBBox());
	((Line*)(markers.current()))->setP1(map(p1,sys,REF));
	((Line*)(markers.current()))->setP2(map(p2,sys,REF));
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerLineArrowCmd(int id, int p1, int p2)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	((Line*)(markers.current()))->setP1Arrow(p1);
	((Line*)(markers.current()))->setP2Arrow(p2);
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerListCmd(MarkerFormat type, CoordSystem sys, 
			  CoordFormat format, unsigned short prop, int which)
{
  char buf[1024];

  switch (type) {
  case PROS:
    int first=1;
    if (markers.head())
      do
	if (!prop || (markers.current()->getProperty(prop) == which)) {
	  ostrstream str(buf,1024);
	  markers.current()->listPros(str, sys, format);
	  if (!first)
	    str << ';' << ends;
	  else {
	    str << ends;
	    first = 0;
	  }
	  Tcl_AppendResult(interp, buf, NULL);
	}
      while (markers.next());
  }
}

void Frame::markerListCmd(MarkerFormat type, CoordSystem sys, 
			  CoordFormat format)
{
  char buf[1024];

  switch (type) {
  case DS9:
    {
      ostrstream str(buf,1024);
      str << "# Region file format: DS9" << endl;
      listCoordSystem(str, sys);
      markerListDefaults(str);
      str << ends;
      Tcl_AppendResult(interp, buf, NULL);
    }
    if (markers.head())
      do {
	ostrstream str(buf,1024);
	markers.current()->list(str, sys, format);
	str << endl << ends;
	Tcl_AppendResult(interp, buf, NULL);
      } while (markers.next());
    return;
  case SAOTNG:
    {
      ostrstream str(buf,1024);
      markerListSAOtngHeader(str, sys, format);
      str << ends;
      Tcl_AppendResult(interp, buf, NULL);
    }
    if (markers.head())
      do {
	ostrstream str(buf,1024);
	markers.current()->listSAOtng(str, sys, format);
	str << endl << ends;
	Tcl_AppendResult(interp, buf, NULL);
      } while (markers.next());
    return;
  case SAOIMAGE:
    if (markers.head())
      do {
	ostrstream str(buf,1024);
	markers.current()->listSAOimage(str);
	str << endl << ends;
	Tcl_AppendResult(interp, buf, NULL);
      } while (markers.next());
    return;
  case PROS:
    if (markers.head())
      do {
	ostrstream str(buf,1024);
	markers.current()->listPros(str, sys, format);
	str << endl << ends;
	Tcl_AppendResult(interp, buf, NULL);
      } while (markers.next());
    return;
  }
}

void Frame::markerListDefaults(ostream& str)
{
  str << endl
      << "color=green" << endl
      << "font=\"system 10 normal\"" << endl
      << "edit=1 move=1 delete=1 include=1 fixed=0" << endl;
}

void Frame::markerListSAOtngHeader(ostream& str, CoordSystem sys, 
				   CoordFormat format)
{
  if (currentFits)
    str << "# filename: " << currentFits->getFileName() << endl;

  switch (sys) {
  case PHYSICAL:
    {
      str << "# format: pixels (physical)" << endl;
    }
    break;
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case ALTAZ:
  case LINEAR:
  case ICRS:
    {
      str << "# format: ";
      switch (format) {
      case DEGREES:
	str << "degrees (";
	break;
      case HMS:
	str << "hms (";
	break;
      }
      listCoordSystem(str, sys);
      str << ')' << endl;
    }
  }
}

void Frame::markerLoadCmd(const char* fn)
{
  ifstream str(fn);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  

  parseMarker(str);
}

void Frame::markerLoadCmd(int fd)
{
  ifstream str(fd);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  

  parseMarker(str);
}

void Frame::markerLoadFitsCmd(const char* fn)
{
  // verify that we have an ext specified
  if (fn && (fn[strlen(fn)-1] != ']')) {
    result = TCL_ERROR;
    return;
  }

  FitsFile* mk = new FitsFitsMMap(fn, FitsFile::EXACT);
  if (!mk->isValid() || !mk->isTable()) {
    delete mk;
    result = TCL_ERROR;
  }

  FitsHead* mkh = mk->head();
  if (mkh->naxes() != 2) {
    delete mk;
    result = TCL_ERROR;
  }
  FitsTableHDU* mkhdu = (FitsTableHDU*)mkh->hdu();

  FitsBinColumnA* shape = (FitsBinColumnA*)mkhdu->find("shape");
  FitsColumn* x = mkhdu->find("x");
  FitsColumn* y = mkhdu->find("y");
  FitsColumn* r = mkhdu->find("r");
  FitsColumn* ang = mkhdu->find("rotang");
  
  if (!shape || !x || !y || !r || !ang) {
    delete mk;
    result = TCL_ERROR;
    return;
  }

  char* ptr = (char*)mk->data();
  int rows = mkhdu->rows();
  int rowlen = mkhdu->width();

  char color[] = "green";
  char font[] = "system 10 normal";
  char text[] = "";
  unsigned short props = Marker::SELECT | Marker::EDIT | Marker::MOVE |
    Marker::ROTATE | Marker::DELETE | Marker::INCLUDE;

  for (int i=0; i<rows; i++, ptr+=rowlen) {
    char* s1 = shape->str(ptr);
    char* s2 = toUpper(s1);
    if (!strncmp(s2, "CIRCLE", 6))
      createCircleCmd(Vector(x->value(ptr,0), y->value(ptr,0)), PHYSICAL,
		      r->value(ptr), PIXELS, color, font, text, props);

    else if (!strncmp(s2, "BOX   ", 6))
      createRectangleCmd(Vector(x->value(ptr,0), y->value(ptr,0)), PHYSICAL,
			 Vector(r->value(ptr,0), r->value(ptr,1)), PIXELS, 
			 0, color, font, text, props);

    else if (!strncmp(s2, "ROTBOX", 6))
      createRectangleCmd(Vector(x->value(ptr,0), y->value(ptr,0)), PHYSICAL,
			 Vector(r->value(ptr,0), r->value(ptr,1)), PIXELS, 
			 degToRad(ang->value(ptr)), color, font, text, props);

    else if (!strncmp(s2, "RECTAN", 6)) {
      Vector v1(x->value(ptr,0), x->value(ptr,1));
      Vector v2(y->value(ptr,0), y->value(ptr,1));
      Vector d = v2-v1;
      Vector c = d/2 + v1;

      createRectangleCmd(c, PHYSICAL, d, PIXELS, 0, color, font, text, props);
    }

    else if (!strncmp(s2, "ROTREC", 6)) {
      Vector v1(x->value(ptr,0), x->value(ptr,1));
      Vector v2(y->value(ptr,0), y->value(ptr,1));
      Vector d = v2-v1;
      Vector c = d/2 + v1;

      createRectangleCmd(c, PHYSICAL, d, PIXELS, degToRad(ang->value(ptr)), 
			 color, font, text, props);
    }

    else if (!strncmp(s2, "ELLIPS", 6))
      createEllipseCmd(Vector(x->value(ptr,0), y->value(ptr,0)), PHYSICAL,
		       Vector(r->value(ptr,0), r->value(ptr,1)), PIXELS, 
		       degToRad(ang->value(ptr)), color, font, text, props);

    else if (!strncmp(s2, "POINT ", 6))
      createBoxCirclePointCmd(Vector(x->value(ptr,0), y->value(ptr,0)),
			      PHYSICAL, color, font, text, props);

    else if (!strncmp(s2, "POLYGO", 6)) {
      List<Vertex>* list = new List<Vertex>();
      Vertex* first = new Vertex(x->value(ptr,0), y->value(ptr,0));
      list->append(first);
      for (int i=1; i<12; i++) {
	Vertex* n = new Vertex(x->value(ptr,i), y->value(ptr,i));
	if (n->vector[0] != first->vector[0] || 
	    n->vector[1] != first->vector[1])
	  list->append(new Vertex(x->value(ptr,i), y->value(ptr,i)));
	else
	  break;
      }

      createPolygonCmd(list, PHYSICAL, color, font, text, props);
    }

    delete s1;
    delete s2;
  }

  delete mk;
}

void Frame::markerMoveCmd(const Vector& v)
{
  if (markers.head())
    do
      if (markers.current()->isSelected() && markers.current()->canMove()) {
	Vector c = markers.current()->getCenter() * refToCanvas;
	update(PIXMAP, markers.current()->getAllBBox());
	markers.current()->moveTo((c + v) * canvasToRef);
	update(PIXMAP, markers.current()->getAllBBox());
      }
    while (markers.next());
}

void Frame::markerMoveCmd(int id, const Vector& v)
{
  if (markers.head())
    do
      if (markers.current()->getId() && markers.current()->canMove()) {
	Vector c = markers.current()->getCenter() * refToCanvas;
	update(PIXMAP, markers.current()->getAllBBox());
	markers.current()->moveTo((c + v) * canvasToRef);
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());
}

void Frame::markerMoveBeginCmd(const Vector& v)
{
  markerBegin = v * canvasToRef;
}

void Frame::markerMoveBeginCmd(int id, const Vector& v)
{
  markerBegin = v * canvasToRef;
}

void Frame::markerMoveMotionCmd(const Vector& v)
{
  // first, accumulate erase markers

  if (markers.head()) {
    do
      if (markers.current()->isSelected() && markers.current()->canMove())
	redraw(markers.current()->getBBox());
    while (markers.next());

    // and erase now

    forceUpdate();

    // ok, now draw selected markers in new location

    Vector markerEnd = v * canvasToRef;
    Vector diff = markerEnd - markerBegin;
    markerBegin = markerEnd;

    markers.head();
    do
      if (markers.current()->isSelected() && markers.current()->canMove()) {
	markers.current()->move(diff);
	markers.current()->drawXOR();
      }
    while (markers.next());
  }
}

void Frame::markerMoveMotionCmd(int id, const Vector& v)
{
  // first, accumulate erase markers

  if (markers.head()) {
    do
      if (markers.current()->getId() && markers.current()->canMove()) {
	redraw(markers.current()->getBBox());
	break;
      }
    while (markers.next());

    if (!markers.current())
      return; // can't find it

    Marker *ptr = markers.current();

    // and erase now

    forceUpdate();

    // ok, now draw selected markers in new location

    Vector markerEnd = v * canvasToRef;
    Vector diff = markerEnd - markerBegin;
    markerBegin = markerEnd;

    ptr->move(diff);
    ptr->drawXOR();
  }
}

void Frame::markerMoveEndCmd()
{
  // update widget since we don't know where the selected markers came from

  update(PIXMAP);
}

void Frame::markerMoveEndCmd(int id)
{
  // update widget since we don't know where the selected markers came from

  update(PIXMAP);
}

void Frame::markerMoveToCmd(const Vector& v, CoordSystem sys)
{
  if (markers.head()) {
    do
      if (markers.current()->isSelected() && markers.current()->canMove()) {
	update(PIXMAP, markers.current()->getAllBBox());
	markers.current()->moveTo(map(v,sys,REF));
	update(PIXMAP, markers.current()->getAllBBox());
      }
    while (markers.next());
  }
}

void Frame::markerMoveToCmd(int id, const Vector& v, CoordSystem sys)
{
  if (markers.head()) {
    do
      if ((markers.current()->getId() == id) && markers.current()->canMove()) {
	update(PIXMAP, markers.current()->getAllBBox());
	markers.current()->moveTo(map(v,sys,REF));
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());
  }
}

void Frame::markerPolygonAngleCmd(int id, double angle)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// it may shrink
	update(PIXMAP, markers.current()->getAllBBox());
	((Polygon*)(markers.current()))->setAngle(angle);
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerPolygonCreateVertexCmd(int id, int seg, const Vector& v)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	((Polygon*)(markers.current()))->createVertex(seg, v * canvasToRef);
	update(PIXMAP, markers.current()->getBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerPolygonDeleteVertexCmd(int id, int h)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	if (markers.current()->canEdit()) {
	  // it may shrink
	  update(PIXMAP, markers.current()->getAllBBox());
	  ((Polygon*)(markers.current()))->deleteVertex(h);
	  update(PIXMAP, markers.current()->getAllBBox());
	}
	return;
      }
    while (markers.next());
}

void Frame::markerPropertyCmd(unsigned short prop, int value)
{
  if (markers.head())
    do {
      Marker* m = markers.current();
      if (m->isSelected()) {
	if (prop == Marker::FIXED || prop == Marker::INCLUDE) {
	  // marker will change bbox, so get before and after
	  update(PIXMAP, m->getAllBBox());
	  m->setProperty(prop, value);
	  update(PIXMAP, m->getAllBBox());
	}
	else
	  m->setProperty(prop, value);
      }
    } while (markers.next());
}

void Frame::markerPropertyCmd(int id, unsigned short prop, int value)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	if (prop == Marker::FIXED || prop == Marker::INCLUDE) {
	  // marker will change bbox, so get before and after
	  update(PIXMAP, markers.current()->getAllBBox());
	  markers.current()->setProperty(prop, value);
	  update(PIXMAP, markers.current()->getAllBBox());
	}
	else
	  markers.current()->setProperty(prop, value);

	return;
      }
    while (markers.next());
}

void Frame::markerRectangleSizeCmd(int id, const Vector& size, 
				   CoordSystem sys, CoordFormat format)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// it may shrink
	update(PIXMAP, markers.current()->getAllBBox());
	((Rectangle*)(markers.current()))->setSize(mapDistToRef(size,sys,
								format));
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerRectangleAngleCmd(int id, double angle)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// it may shrink
	update(PIXMAP, markers.current()->getAllBBox());
	((Rectangle*)(markers.current()))->setAngle(angle);
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerRotateBeginCmd(int id)
{
  // remember which marker is being edited

  if (markers.head())
    do
      if (markers.current()->getId() == id && markers.current()->canRotate()) {
	rotateMarker = markers.current();
	return;
      }
    while (markers.next());

  rotateMarker = NULL;
}

void Frame::markerRotateBeginCmd(const Vector& v)
{
  // remember which marker is being edited

  if (markers.head())
    do
      if (markers.current()->isSelected() && markers.current()->canRotate()) {
	rotateMarker = markers.current();
	return;
      }
    while (markers.next());

  rotateMarker = NULL;
}

void Frame::markerRotateMotionCmd(const Vector& v, int h)
{
  if (rotateMarker) {

    // erase current marker now

    redraw(rotateMarker->getBBox());
    forceUpdate();

    rotateMarker->rotate(v*canvasToRef, h);
    rotateMarker->drawXOR();
  }
}

void Frame::markerRotateEndCmd()
{
  rotateMarker = NULL;
  update(PIXMAP);
}

void Frame::markerRulerPointCmd(int id, const Vector& p1, const Vector& p2,
				CoordSystem sys)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// it may shrink
	update(PIXMAP, markers.current()->getAllBBox());
	((Ruler*)(markers.current()))->setP1(map(p1,sys,REF));
	((Ruler*)(markers.current()))->setP2(map(p2,sys,REF));
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerRulerSystemCmd(int id, CoordSystem sys, CoordFormat format)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// it may shrink
	update(PIXMAP, markers.current()->getAllBBox());
	((Ruler*)(markers.current()))->setSystem(sys, format);
	update(PIXMAP, markers.current()->getAllBBox());
	return;
      }
    while (markers.next());

  result = TCL_ERROR;
}

void Frame::markerSaveCmd(const char* fileName, MarkerFormat type, 
			  CoordSystem sys, CoordFormat format)
{
  ofstream fn(fileName);
  if (fn) {
    switch (type) {
    case DS9:
      listCoordSystem(fn, sys);
      markerListDefaults(fn);
      if (markers.head())
	do {
	  markers.current()->list(fn, sys, format);
	  fn << endl;
	}
	while (markers.next());
      break;
    case SAOTNG:
      markerListSAOtngHeader(fn, sys, format);
      if (markers.head())
	do {
	  markers.current()->listSAOtng(fn, sys, format);
	  fn << endl;
	}
	while (markers.next());
      return;
    case SAOIMAGE:
      if (markers.head())
	do {
	  markers.current()->listSAOimage(fn);
	  fn << endl;
	}
	while (markers.next());
      return;
    case PROS:
    if (markers.head())
      do {
	markers.current()->listPros(fn, sys, format);
	fn << endl;
      }
      while (markers.next());
    return;
    }

  }
  else {
    Tcl_AppendResult(interp, "Unable to open file ", fileName, NULL);
    result = TCL_ERROR;
  }
}

void Frame::markerSelectCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	markers.current()->select();
	update(PIXMAP, markers.current()->getBBox());
	return;
      }
    while (markers.next());
}

void Frame::markerSelectCmd(const Vector& v)
{
  // first, check to see if we clicked on an already selected marker

  if (markers.head())
    do
      if (markers.current()->isIn(v) && markers.current()->isSelected()) {
	Tcl_AppendResult(interp, "1", NULL);
	return;
      }
    while (markers.next());

  // ok, now select the first found, and unselect the rest

  int found = 0;
  if (markers.head())
    do {
      Marker* m = markers.current();
      if (m->isIn(v) && m->canSelect() && !found) {
	if (!m->isSelected()) {
	  m->select();
	  update(PIXMAP, m->getBBox());
	}
	found = 1;
      }
      else {
	if (m->isSelected()) {
	  m->unselect();
	  update(PIXMAP, m->getBBox());
	}
      }
    } while (markers.next());

  if (found)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::markerSelectAllCmd()
{
  if (markers.head())
    do
      if (markers.current()->canSelect()) {
	markers.current()->select();
	update(PIXMAP, markers.current()->getBBox());
      }
    while (markers.next());
}

void Frame::markerSelectToggleCmd(const Vector& v)
{
  // toggle the select of the first found

  if (markers.head())
    do
      if (markers.current()->isIn(v) && markers.current()->canSelect()) {
	markers.current()->toggleSelect();
	update(PIXMAP, markers.current()->getBBox());
	Tcl_AppendResult(interp, "1", NULL);

	return;
      }
    while (markers.next());

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::markerTextCmd(int id, const char* text)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	// things can shrink, so do before and after
	  update(PIXMAP,markers.current()->getAllBBox());
	  markers.current()->setText(text);
	  update(PIXMAP,markers.current()->getAllBBox());
	  
	  return;
      }
    while (markers.next());
}

void Frame::markerUnselectCmd(int id)
{
  if (markers.head())
    do
      if (markers.current()->getId() == id) {
	markers.current()->unselect();
	update(PIXMAP, markers.current()->getBBox());
	return;
      }
    while (markers.next());
}

void Frame::markerUnselectAllCmd()
{
  if (markers.head())
    do {
      markers.current()->unselect();
      update(PIXMAP, markers.current()->getBBox());
    } while (markers.next());
}

void Frame::regionSelectBeginCmd(const Vector& v)
{
  regionBegin = v;
  regionEnd = v;
}

void Frame::regionSelectMotionCmd(const Vector& v)
{
  if (visible) {

    // erase 

    redrawNow((BBox(regionBegin, regionEnd)).expand(2));

    // and draw to window

    regionEnd = v;
    BBox cc = BBox(regionBegin, regionEnd) * canvasToWindow;
    Vector size = cc.size();

    XDrawRectangle(display, Tk_WindowId(tkwin), markerGCXOR,
		   (int)cc.ll[0], (int)cc.ll[1], (int)size[0], (int)size[1]);
  }
}

void Frame::regionSelectEndCmd()
{
  BBox bb(regionBegin, regionEnd);

  if (markers.head())
    do {
      Marker* m = markers.current();
      if (bb.isIn(m->getBBox())==4 && m->canSelect())
	m->select();
      else
	m->unselect();
    } while (markers.next());

  update(PIXMAP, bb.expand(2));
}

void Frame::regionSelectShiftEndCmd()
{
  BBox bb(regionBegin, regionEnd);

  if (markers.head())
    do {
      Marker* m = markers.current();
      if (bb.isIn(m->getBBox())==4)
	m->select();
    } while (markers.next());

  update(PIXMAP, bb.expand(2));
}

// Marker Support

void Frame::parseMarker(istream& str)
{
  mklexx = new mkFlexLexer(&str);
  mkparse(this);
  delete mklexx;
  mklexx = NULL;
}

void Frame::psMarkers(int mode)
{
  // render from back to front
  // bbox is in canvas coords

  const BBox bb = BBox(0, 0, options->width-1, options->height-1) * 
    widgetToCanvas;

  if (markers.tail())
    do
      if (markers.current()->isVisible(bb))
	markers.current()->ps(mode);
    while (markers.previous());
}

void Frame::renderMarkers(const BBox& bb)
{
  // render from back to front
  // bbox is in canvas coords

  if (markers.tail())
    do
      if (markers.current()->isVisible(bb))
	markers.current()->draw();
    while (markers.previous());
}

void Frame::renderMagnifierMarkers(const Matrix& matrix, const BBox& bb)
{
  // render from back to front
  // bbox is in canvas coords

  if (markers.tail())
    do
      if (markers.current()->isVisible(bb))
	markers.current()->drawMagnifier(matrix);
    while (markers.previous());
}

void Frame::unselectMarkers()
{
  if (markers.head())
    do
      markers.current()->unselect();
    while (markers.next());
}

void Frame::updateMarkers()
{
  if (markers.head())
    do
      markers.current()->updateBBox();
    while (markers.next());
}

void Frame::updateMarkerCoords(Matrix& m)
{
  if (markers.head())
    do
      markers.current()->updateCoords(m);
    while (markers.next());
}
