%{
#define YYPARSE_PARAM fr
#define FR ((Frame*)fr)
#define YYDEBUG 1

#define DISCARD_(x) {yyclearin; mkDiscard(x);}

#include <math.h>
#include <string.h>

#include <iostream.h>

#include "frame.h"
#include "util.h"
#include "vector.h"
#include "list.h"

extern int mklex(void);
extern void mkerror(const char*);
extern void mkDiscard(int);

static Frame::CoordSystem globalSystem;
static Frame::CoordSystem currentSystem;

static Frame::CoordFormat currentFormat;

static unsigned short globalProps;
static unsigned short currentProps;

static char globalColor[16];
static char currentColor[16];

static char globalFont[32];
static char currentFont[32];

static char globalText[80];
static char currentText[80];

static Vector currentVector;
static Vector currentVector2;

static int arrow1;
static int arrow2;

static List<Vertex>* polylist;

static int tngText;
static int tngHeader;

static void setProps(unsigned short* props, unsigned short prop, int value);
static void setWCSSystem();
	
%}

%union {
  float real;
  int integer;
  char str[256];
}

%type <real> numeric
%type <integer> yesno

%type <real> angle
%type <real> value
%type <real> hmsCoord
%type <real> dmsCoord
%type <integer> property
%type <integer> coordSystem
%type <integer> coordFormat
%type <integer> wcsSystem
%type <integer> equatorial

%token <integer> INT
%token <real> REAL
%token <str> STRING

%token <integer> HOUR
%token <integer> MINUTE
%token <real> SECOND

%token <real> DEGREE
%token <real> ARCMINUTE
%token <real> ARCSECOND

%token <str> HMSSTR

%token <real> RADIAN
%token <integer> PIXEL

%token AMPLIFIER_
%token ANNULUS_
%token ARCMIN_
%token ARCSEC_
%token ARROW_
%token B1950_
%token BOX_
%token BOXCIRCLE_
%token CIRCLE_
%token COLOR_
%token CROSS_
%token DATE_
%token DEBUG_
%token DEGREES_
%token DELETE_
%token DETECTOR_
%token DIAMOND_
%token ECLIPTIC_
%token EDIT_
%token ELLIPSE_
%token EQUATORIAL_
%token FALSE_
%token FILENAME_
%token FIXED_
%token FK4_
%token FK5_
%token FONT_
%token FORMAT_
%token GALACTIC_
%token GLOBAL_
%token HMS_
%token ICRS_
%token IMAGE_
%token INCLUDE_
%token J2000_
%token LINE_
%token LINEAR_
%token LOGICAL_
%token MOVE_
%token N_
%token NO_
%token OFF_
%token ON_
%token PHYSICAL_
%token PIXELS_
%token POINT_
%token POLYGON_
%token RECTANGLE_
%token ROTATE_
%token ROTBOX_
%token RULER_
%token SELECT_
%token TEXT_
%token TRUE_
%token VERSION_
%token X_
%token Y_
%token YES_

%%

start	: {
	  // init globals
	  globalSystem = Frame::IMAGE;
	  globalProps = Marker::SELECT | Marker::EDIT | Marker::MOVE |
	    Marker::ROTATE | Marker::DELETE | Marker::INCLUDE;
	  strcpy(globalColor,"green");
	  strcpy(globalFont,"system 10 normal");
	  strcpy(globalText,"");
	  tngText = 0;   // we have found a comment to be used as text
	  tngHeader = 0; // we have a tng header
	} commands
	;

commands: commands command terminator
	| command terminator
	;

command : /* empty */
	| DEBUG_ debug
	| VERSION_ {cerr << "Regions File 1.0" << endl;}
	| globalProperties eolComment
	| coordSystem {globalSystem = (Frame::CoordSystem)$1;} eolComment
	| initProps include shape
	| comment
	;

terminator: '\n'
	| ';'
	;

numeric	: REAL {$$=$1;}
	| INT {$$=$1;}
	;

debug	: ON_ {yydebug=1;}
	| OFF_ {yydebug=0;}
	;

yesno	: INT {$$=($1 ? 1 : 0);}

	| YES_ {$$=1;}
	| Y_ {$$=1;}
	| ON_ {$$=1;}
	| TRUE_ {$$=1;}

	| NO_ {$$=0;}
	| N_ {$$=0;}
	| OFF_ {$$=0;}
	| FALSE_ {$$=0;}
	;

sp	: /* empty */
	| ','
	;

bp	: /* empty */
	| '('
	;

ep	: /* emtpy */
	| ')'
	;

angle	: /* empty */ {$$ = 0;}
	| numeric {$$ = degToRad($1);} /* assume degree */
	| DEGREE {$$ = degToRad($1);} /* assume degree */
	| RADIAN {$$=$1;}
	;

value	: numeric {$$=$1; currentFormat = Frame::PIXELS;}
	| PIXEL   {$$=$1; currentFormat = Frame::PIXELS;}
	| DEGREE  {$$=$1; currentFormat = Frame::DEGREES;}
	| ARCMINUTE {$$=$1; currentFormat = Frame::ARCMIN;}
	| ARCSECOND {$$=$1; currentFormat = Frame::ARCSEC;}
	;

coord	: numeric sp numeric {currentVector = Vector($1,$3);}
	| PIXEL sp PIXEL {currentVector = Vector($1,$3);}
	| DEGREE sp DEGREE {setWCSSystem(); currentVector = Vector($1,$3);}
	| hmsCoord sp dmsCoord {setWCSSystem(); currentVector = Vector($1,$3);}
	;

hmsCoord : HMSSTR {$$=parseHMS($1);}
	| HOUR MINUTE SECOND {$$=hmsToDegree(hmsSign,$1,$2,$3);}
	;

dmsCoord : HMSSTR {$$=parseDMS($1);}
	| numeric {hmsSign = ($1>=0?1:-1);} ARCMINUTE ARCSECOND 
	  {$$=dmsToDegree(hmsSign,fabs($1),$3,$4);}
	;

property : SELECT_ {$$ = Marker::SELECT;}
	| EDIT_ {$$ = Marker::EDIT;}
	| MOVE_  {$$ = Marker::MOVE;}
	| ROTATE_ {$$ = Marker::ROTATE;}
	| DELETE_ {$$ = Marker::DELETE;}
	| INCLUDE_ {$$ = Marker::INCLUDE;}
	| FIXED_ {$$ = Marker::FIXED;}
	;

coordSystem : IMAGE_ {$$ = Frame::IMAGE;}
	| LOGICAL_ {$$ = Frame::IMAGE;}
	| PHYSICAL_ {$$ = Frame::PHYSICAL;}
	| AMPLIFIER_ {$$ = Frame::AMPLIFIER;}
	| DETECTOR_ {$$ = Frame::DETECTOR;}
	| wcsSystem {$$ = $1;}
	;

coordFormat : /* empty */ {$$=Frame::PIXELS;}
	| PIXELS_ {$$=Frame::PIXELS;}
	| DEGREES_ {$$=Frame::DEGREES;}
	| ARCMIN_ {$$=Frame::ARCMIN;}
	| ARCSEC_ {$$=Frame::ARCSEC;}
	;

wcsSystem : EQUATORIAL_ equatorial {$$ = $2;}
	| FK4_ {$$ = Frame::EQUATORIAL_FK4;}
	| FK5_ {$$ = Frame::EQUATORIAL_FK5;}
	| B1950_  {$$ = Frame::EQUATORIAL_FK4;}
	| J2000_  {$$ = Frame::EQUATORIAL_FK5;}	
	| GALACTIC_ {$$ = Frame::GALACTIC;}
	| ECLIPTIC_ {$$ = Frame::ECLIPTIC;}
	| LINEAR_ {$$ = Frame::LINEAR;}
	| ICRS_ {$$ = Frame::ICRS;}
	;

equatorial : B1950_  {$$ = Frame::EQUATORIAL_FK4;}
	| J2000_  {$$ = Frame::EQUATORIAL_FK5;}
	| FK4_  {$$ = Frame::EQUATORIAL_FK4;}
	| FK5_  {$$ = Frame::EQUATORIAL_FK5;}
	;

initProps : {
	  currentProps = globalProps;
	  currentSystem = globalSystem;
	  strcpy(currentColor,globalColor);
	  strcpy(currentFont,globalFont);
	  strcpy(currentText,globalText);
	  /* we need this for tng format text comments */
	  if (tngText) {
	    strcpy(globalText,"");
	    tngText=0;
          }
	}
	;

globalProperties : globalProperties sp globalProperty
	| globalProperty
	;

globalProperty : property '=' yesno {setProps(&globalProps,$1,$3);}
	| COLOR_ '=' STRING {strncpy(globalColor,$3,16);}
	| FONT_ '=' STRING {strncpy(globalFont,$3,32);}
	| GLOBAL_ TEXT_ '=' STRING {strncpy(globalText,$4,80);}
	;

markerCoordSystem : /* empty */
	| coordSystem {currentSystem = (Frame::CoordSystem)$1;}
	;

markerProperties : /* empty */
	| markerProps
	;

markerProps : markerProps sp markerProp
	| markerProp
	;

markerProp : property '=' yesno {setProps(&currentProps,$1,$3);}
	| COLOR_ '=' STRING {strncpy(currentColor,$3,16);}
	| FONT_ '=' STRING {strncpy(currentFont,$3,32);}
	| TEXT_ '=' STRING {strncpy(currentText,$3,80);}
	;

arrows	: /* empty */ {arrow1=0; arrow2=0;}
	| ARROW_ INT INT {arrow1=$2; arrow2=$3;}
	;

include	: /* empty */
	| '+' {setProps(&currentProps, Marker::INCLUDE, 1);}
	| '-' {setProps(&currentProps, Marker::INCLUDE, 0);}
	;

shape	: CIRCLE_ bp coord sp value ep markerCoordSystem 
	    markerProperties shapeComment
	  {FR->createCircleCmd(currentVector,currentSystem,$5,currentFormat,
	    currentColor,currentFont,currentText,currentProps);}
	| ELLIPSE_ bp coord sp value sp value sp angle ep 
	    markerCoordSystem markerProperties shapeComment
	  {FR->createEllipseCmd(currentVector,currentSystem,Vector($5,$7),
	    currentFormat,$9,currentColor,currentFont,currentText,
	    currentProps);}
	| RECTANGLE_ bp coord sp value sp value sp angle ep 
	    markerCoordSystem markerProperties shapeComment
	  {FR->createRectangleCmd(currentVector,currentSystem,Vector($5,$7),
	    currentFormat,$9,currentColor,currentFont,currentText,
	    currentProps);}
	| POLYGON_ {polylist=new List<Vertex>();} 
	    bp polyNodes ep markerCoordSystem markerProperties shapeComment
	  {FR->createPolygonCmd(polylist, currentSystem,
	    currentColor,currentFont,currentText,currentProps);} 
	| LINE_ bp coord {currentVector2 = currentVector;}
	    sp coord ep markerCoordSystem markerProperties arrows shapeComment
	  {FR->createLineCmd(currentVector2,currentVector,currentSystem,
	    arrow1,arrow2,currentColor,currentFont,currentText,currentProps);}
	| TEXT_ bp coord ep markerCoordSystem markerProperties shapeComment
	  {FR->createTextCmd(currentVector,currentSystem,
	    currentColor,currentFont,currentText,currentProps);}
	| RULER_ bp coord {currentVector2 = currentVector;} 
	    sp coord ep markerCoordSystem coordFormat markerProperties
	    shapeComment
	  {FR->createRulerCmd(currentVector2,currentVector,currentSystem,
	    (Frame::CoordFormat)$9,
	    currentColor,currentFont,currentText,currentProps);}
	| CIRCLE_ POINT_ bp coord ep markerCoordSystem markerProperties 
            shapeComment
	  {FR->createCirclePointCmd(currentVector,currentSystem,
	    currentColor,currentFont,currentText,currentProps);}
	| BOX_ POINT_ bp coord ep markerCoordSystem markerProperties
	    shapeComment
	  {FR->createBoxPointCmd(currentVector,currentSystem,
	    currentColor,currentFont,currentText,currentProps);}
	| DIAMOND_ POINT_ bp coord ep markerCoordSystem markerProperties
	    shapeComment
	  {FR->createDiamondPointCmd(currentVector,currentSystem,
	    currentColor,currentFont,currentText,currentProps);}
	| CROSS_ POINT_ bp coord ep markerCoordSystem markerProperties
	     shapeComment
	  {FR->createCrossPointCmd(currentVector,currentSystem,
	    currentColor,currentFont,currentText,currentProps);}
	| X_ POINT_ bp coord ep markerCoordSystem markerProperties shapeComment
	  {FR->createExPointCmd(currentVector,currentSystem,
	    currentColor,currentFont,currentText,currentProps);}
	| ARROW_ POINT_ bp coord ep markerCoordSystem markerProperties
	     shapeComment
	  {FR->createArrowPointCmd(currentVector,currentSystem,
	    currentColor,currentFont,currentText,currentProps);}
	| BOXCIRCLE_ POINT_ bp coord ep markerCoordSystem 
	    markerProperties shapeComment
	  {FR->createBoxCirclePointCmd(currentVector,currentSystem,
	    currentColor,currentFont,currentText,currentProps);}
	| POINT_ bp coord ep markerCoordSystem markerProperties shapeComment
	  {FR->createBoxCirclePointCmd(currentVector,currentSystem,
	    currentColor,currentFont,currentText,currentProps);}

	| BOX_ bp coord sp numeric sp numeric sp angle ep 
	    markerCoordSystem markerProperties shapeComment
	  {FR->createRectangleCmd(currentVector,currentSystem,Vector($5,$7),
	    Frame::PIXELS,$9,currentColor,currentFont,currentText,
	    currentProps);}
	| TEXT_ bp coord sp STRING ep {strncpy(currentText,$5,80);}
	    markerCoordSystem markerProperties shapeComment
	  {FR->createTextCmd(currentVector,currentSystem,
	    currentColor,currentFont,currentText,currentProps);}

	| ROTBOX_ bp coord sp value sp value sp angle ep 
	    markerCoordSystem markerProperties shapeComment
	  {FR->createRectangleCmd(currentVector,currentSystem,Vector($5,$7),
	    currentFormat,$9,currentColor,currentFont,currentText,
	    currentProps);}
	;

polyNodes : polyNodes sp polyNode
	| polyNode
	;

polyNode : coord {polylist->append(new Vertex(currentVector));}
	;

comment	: '#' FORMAT_ ':' {tngHeader=1;} tngFormat
	| '#' FILENAME_ ':' {tngHeader=1; DISCARD_(0)} STRING
	| '#' DATE_ ':' {tngHeader=1; DISCARD_(0)} STRING
	| '#' {DISCARD_(1)} STRING 
	{
	  /* we need this for tng format text comments */
	  if (tngHeader) {
	    strncpy(globalText,$3,80);
	    tngText=1;
	  }
	}
	| '#' ARROW_ {DISCARD_(0)} STRING
	;

eolComment : /* empty */
	| '#' {DISCARD_(0)} STRING
	;

shapeComment : /* empty */
	| '#' {DISCARD_(1)} STRING
	{
	  /* try to feature out if this is a color, not a comment */
	  if (!strcmp($3,"# white"))
	    strcpy(currentColor,"white");
	  else if (!strcmp($3,"# black"))
	    strcpy(currentColor,"black");
	  else if (!strcmp($3,"# red"))
	    strcpy(currentColor,"red");
	  else if (!strcmp($3,"# green"))
	    strcpy(currentColor,"green");
	  else if (!strcmp($3,"# blue"))
	    strcpy(currentColor,"blue");
	  else if (!strcmp($3,"# cyan"))
	    strcpy(currentColor,"cyan");
	  else if (!strcmp($3,"# magneta"))
	    strcpy(currentColor,"magneta");
	  else if (!strcmp($3,"# yellow"))
	    strcpy(currentColor,"yellow");

	  else if (!strcmp($3,"# source"))
	    strcpy(currentColor,"green");
	  else if (!strcmp($3,"# background"))
	    strcpy(currentColor,"red");

	  else if (!strcmp($3,"# purple"))
	    strcpy(currentColor,"magneta");
	  else if (!strcmp($3,"# darkslategray"))
	    strcpy(currentColor,"white");
        }
	;

tngFormat : HMS_ tngWCS
	| DEGREES_ tngWCS
	| PIXELS_ {globalSystem = Frame::IMAGE;}
	| PIXELS_ '(' PHYSICAL_ ')' {globalSystem = Frame::PHYSICAL;}
	;

tngWCS	: /* empty */ {globalSystem = Frame::EQUATORIAL_FK4;}
	| '(' wcsSystem ')' {globalSystem = (Frame::CoordSystem)$2;}
	;

%%

static void setProps(unsigned short* props, unsigned short prop, int value)
{
  if (value)
    *props |= prop;
  else
    *props &= ~prop;
}

static void setWCSSystem()
{
  // if we have not defined a specific WCS, set to Native
  switch (currentSystem) {
    case (Frame::IMAGE):
    case (Frame::PHYSICAL):
    case (Frame::DETECTOR):
    case (Frame::AMPLIFIER):
      currentSystem = (Frame::NATIVEWCS);
      break;
  }
}
